"""
Various retry strategies, from no retry to exponential backoff.

We provide predefined strategies, as well as baseclasses for you to define your own.
Predefined are:
- `NoRetryStrategy`
- `ImmediateRetryStrategy`
- `ExponentialBackoffStrategy`

To define your own strategy, implement your own subclass of `RetryStrategy`,
which returns fresh instances of your own `RetryEvaluator` subclass.
"""

from asyncio import sleep
from abc import ABC, abstractmethod
from datetime import timedelta


class RetryEvaluator(ABC):
    """
    This class performs the actual retry handling.
    It can keep track of internal state as it so wishes,
    to perform the function of `should_retry`.
    """

    @abstractmethod
    async def should_retry(self) -> bool:
        """
        This method should return True if the strategy should retry,
        but may wait internally if it deems necessary.
        """
        pass


class RetryStrategy(ABC):
    """
    Base class for all retry strategies, whether predefined or user-made.
    The strategy has as its sole purpose to create instances of `RetryEvaluator`
    that can be used for a specific operation.

    Each individual failable operation will request a separate evaluator using the make method.
    """

    @abstractmethod
    def make(self) -> RetryEvaluator:
        """
        Create an instance of a RetryEvaluator.
        Custom strategies should also subclass RetryEvaluator
        """
        pass


class NoRetryStrategy(RetryStrategy):
    """
    Retry strategy that simply does not retry.
    """

    def make(self) -> RetryEvaluator:
        class NoRetryEvaluator(RetryEvaluator):
            async def should_retry(self) -> bool:
                return False

        return NoRetryEvaluator()


class ImmediateRetryStrategy(RetryStrategy):
    """
    Retry strategy that tries again without delay,
    up to a user defined maximum amount of times.
    """

    def __init__(self, max_retries: int) -> None:
        self.max_retries = max_retries

    def make(self) -> RetryEvaluator:
        class ImmediateRetryEvaluator(RetryEvaluator):
            count: int = 0
            max_retries: int = self.max_retries

            async def should_retry(self) -> bool:
                if self.count >= self.max_retries:
                    return False
                self.count += 1
                return True

        return ImmediateRetryEvaluator()


class ExponentialBackoffStrategy(RetryStrategy):
    """
    Retry strategy implementing the exponential backoff algorithm.
    Every failure up to `max_retries` will result in a sleep
    of $t^n$ seconds, with $t$ being the backoff and $n$ amount of failures.

    Note that backoff values less than one second will count as zero.
    """

    max_retries: int
    backoff: timedelta
    max_backoff: timedelta

    def __init__(
        self,
        max_retries: int = 5,
        backoff: timedelta = timedelta(seconds=2),
        max_backoff: timedelta = timedelta(hours=24),
    ) -> None:
        self.max_retries = max_retries
        self.backoff = backoff
        self.max_backoff = max_backoff

    def make(self) -> RetryEvaluator:
        class ExponentialBackoffEvaluator(RetryEvaluator):
            count: int = 0
            max_retries: int = self.max_retries
            backoff: timedelta = self.backoff
            max_backoff: timedelta = self.max_backoff

            async def should_retry(self) -> bool:
                if self.count >= self.max_retries:
                    return False
                self.count += 1
                await sleep(
                    min(self.max_backoff.seconds, self.backoff.seconds**self.count)
                )
                return True

        return ExponentialBackoffEvaluator()



from obelisk.types.core import Filter, Comparison
from datetime import datetime


def test_basic_filter():
    test_dt = datetime.now()
    f = (
        Filter()
        .add_and(
            Comparison.equal("source", "test source"),
        )
        .add_or(Comparison.less("timestamp", test_dt))
        .add_or(
            Comparison.is_in("metricType", ["number", "number[]"]),
        )
    )

    expected = f"((('source'=='test source'),'timestamp'<'{test_dt.isoformat()}'),'metricType'=in=('number', 'number[]'))"
    assert str(f) == expected



"""
This module contains the asynchronous API to interface with Obelisk CORE.
These methods all return a `Awaitable`.

Relevant entrance points are `Client`.

This API vaguely resembles that of clients to previous Obelisk versions,
but also significantly diverts from it where the underlying Obelisk CORE API does so.
"""

from obelisk.asynchronous.base import BaseClient
from obelisk.exceptions import ObeliskError
from obelisk.types.core import FieldName, Filter

from datetime import datetime, timedelta
import httpx
import json
from pydantic import (
    BaseModel,
    AwareDatetime,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
    SerializerFunctionWrapHandler,
    WrapSerializer,
    field_serializer,
)
from typing import (
    Annotated,
    Literal,
    Any,
    cast,
    get_args,
)
from collections.abc import AsyncIterator, Iterator
from typing_extensions import Self
from numbers import Number

from obelisk.strategies.retry import NoRetryStrategy, RetryStrategy
from obelisk.types import ObeliskKind
from obelisk.types.core import IngestMode


DataType = Literal["number", "number[]", "json", "bool", "string"]
"""The possible types of data Obelisk can accept"""


def type_suffix(metric: str) -> DataType:
    """
    Extracts the :any:`DataType` from a string metric,
    useful for the dataType field in queries.

    Throws a :py:exc:`ValueError` if the provided string does not appear to be a typed metric,
    or the found type suffix is not a known one.
    """
    split = metric.split("::")

    if len(split) != 2:
        raise ValueError("Incorrect amount of type qualifiers")

    suffix = split[1]
    if suffix not in get_args(DataType):
        raise ValueError(
            f"Invalid type suffix, should be one of {', '.join(get_args(DataType))}"
        )
    return cast(DataType, suffix)


Aggregator = Literal["last", "min", "mean", "max", "count", "stddev"]
"""Type of aggregation Obelisk can process"""


Datapoint = dict[str, Any]
"""Datapoints resulting from queries are modeled as simple dicts, as fields can come and go depending on query."""


class ObeliskPosition(BaseModel):
    """
    Format for coordinates as expected by Obelisk.
    """

    lat: float
    """Latitude"""
    lng: float
    """Longitude"""
    elevation: float


class IncomingDatapoint(BaseModel):
    """A datapoint to be submitted to Obelisk. These are validated quite extensively, but not fully.
    We check roughly if the value type corresponds to the declared type if its one of `number`, `number[]`, `bool` or `string`.
    """

    timestamp: AwareDatetime | None = None
    metric: str
    value: Any
    labels: dict[str, str] | None = None
    location: ObeliskPosition | None = None

    @model_validator(mode="after")
    def check_metric_type(self) -> Self:
        suffix = type_suffix(self.metric)

        if suffix == "number" and not isinstance(self.value, Number):
            raise ValueError(
                f"Type suffix mismatch, expected number, got {type(self.value)}"
            )

        if suffix == "number[]" and (
            type(self.value) is not list
            or any([not isinstance(x, Number) for x in self.value])
        ):
            raise ValueError("Type suffix mismatch, expected value of number[]")

        # Do not check json, most things should be serialisable

        if suffix == "bool" and type(self.value) is not bool:
            raise ValueError(
                f"Type suffix mismatch, expected bool, got {type(self.value)}"
            )

        if suffix == "string" and type(self.value) is not str:
            raise ValueError(
                f"Type suffix mismatch, expected bool, got {type(self.value)}"
            )

        return self


def serialize_comma_string(
    input: Any, handler: SerializerFunctionWrapHandler
) -> str | None:
    if val := handler(input):
        return ",".join(val)
    return None


class QueryParams(BaseModel):
    """
    To avoid having too many parameters on query functions,
    and sharing the implementation between query and chunked query,
    this model collects the information needed to execute a query.

    Contrary to the name, this does not correlate directly to URL query parameters sent to Obelisk.
    """

    dataset: str
    groupBy: Annotated[
        list[FieldName] | None, WrapSerializer(serialize_comma_string)
    ] = None
    """List of Field Names to aggregate by as defined in Obelisk docs, None selects the server-side defaults."""
    aggregator: Aggregator | None = None
    fields: Annotated[
        list[FieldName] | None, WrapSerializer(serialize_comma_string)
    ] = None
    """List of Field Names as defined in Obelisk docs, None selects the server-side defaults."""
    orderBy: Annotated[list[str] | None, WrapSerializer(serialize_comma_string)] = None
    """List of Field Names, with their potential prefixes and suffixes, to select ordering. None user server defaults."""
    dataType: DataType | None = None
    """Data type expected to be returned, is mandatory if the `value` field is requested in the `fields` parameter"""
    filter_: Annotated[str | Filter | None, Field(serialization_alias="filter",)] = None
    """
    Obelisk CORE handles filtering in [RSQL format](https://obelisk.pages.ilabt.imec.be/obelisk-core/query.html#rsql-format),
    to make it easier to also programatically write these filters, we provide the `obelisk.types.core.Filter` option as well.

    Suffix to avoid collisions with builtin Python filter function.
    """
    cursor: str | None = None
    limit: int = 1000

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def check_datatype_needed(self) -> Self:
        if (self.fields is None or "value" in self.fields) and self.dataType is None:
            raise ValueError("Value field requested, must specify datatype")

        return self

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump(
            exclude_none=True, by_alias=True, mode="json", exclude={"dataset"}
        )

    @field_serializer('filter_', mode='plain')
    def serialize_filter(self, value: Filter | str | None) -> str | None:
        if value is None or isinstance(value, str):
            return value

        return str(value)


class ChunkedParams(BaseModel):
    """
    The parameters to be used with `Client.query_time_chunked`,
    which allows fetching large spans of data in specified "chunks" specified in time units,
    for example processing weeks of data one hour at a time.
    This limits memory useage.
    """

    dataset: str
    groupBy: list[FieldName] | None = None
    aggregator: Aggregator | None = None
    fields: list[FieldName] | None = None
    orderBy: list[str] | None = (
        None  # More complex than just FieldName, can be prefixed with - to invert sort
    )
    dataType: DataType | None = None
    filter_: str | Filter | None = None
    """Underscore suffix to avoid name collisions"""
    start: datetime
    end: datetime
    jump: timedelta = timedelta(hours=1)
    """The size of one chunk. 1 hour is a common default. You will receive however many datapoints are included in this interval."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def check_datatype_needed(self) -> Self:
        if (self.fields is None or "value" in self.fields) and self.dataType is None:
            raise ValueError("Value field requested, must specify datatype")

        return self

    def chunks(self) -> Iterator[QueryParams]:
        """Splits this model into an Iterator of ordinary `QueryParams` objects, to query one timestep at a time."""
        current_start = self.start
        while current_start < self.end:
            current_end = current_start + self.jump
            filter_ = f"timestamp>={current_start.isoformat()};timestamp<{current_end.isoformat()}"
            if self.filter_:
                filter_ += f";{self.filter_}"

            yield QueryParams(
                dataset=self.dataset,
                groupBy=self.groupBy,
                aggregator=self.aggregator,
                fields=self.fields,
                orderBy=self.orderBy,
                dataType=self.dataType,
                filter_=filter_,
            )

            current_start += self.jump


class QueryResult(BaseModel):
    """The data returned by a single chunk fetch"""

    cursor: str | None = None
    """Cursors always point to the next page of data matched by filters.
    They are none if there is no more data, they do not consider datapoint count limits."""
    items: list[Datapoint]


class Client(BaseClient):
    """
    This class performs all communication with Obelisk.

    The intended methods to be used by consumers are `query` or `query_time_chunked`.
    These will respectively return all data matching specified parameters,
    or return all data, one timestep at a time respectively.

    `send` is considered an implementation detail,
    but may be used by consumers for any endpoints not yet implemented by obelisk-py.

    `fetch_single_chunk` is the underlying layer to both query methods and requires the user to handle cursors themselves.
    It may however still be useful in some circumstances.
    """

    page_limit: int = 250
    """How many datapoints to request per page in a cursored fetch"""

    def __init__(
        self,
        client: str,
        secret: str,
        retry_strategy: RetryStrategy = NoRetryStrategy(),  # noqa: B008   # This is fine to be shared
    ) -> None:
        BaseClient.__init__(
            self,
            client=client,
            secret=secret,
            retry_strategy=retry_strategy,
            kind=ObeliskKind.CORE,
        )

    async def send(
        self,
        dataset: str,
        data: list[IncomingDatapoint],
        ingest_mode: IngestMode = IngestMode.BOTH,
    ) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        - dataset
            ID for the dataset to publish to
        - data
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.

        Raises
        ------

        - ObeliskError
            When the resulting status code is not 204, an `obelisk.exceptions.ObeliskError` is raised.
        """

        response = await self.http_post(
            f"{self.kind.root_url}/{dataset}/data/ingest",
            data=[x.model_dump(mode="json") for x in data],
            params={"mode": ingest_mode.value}
        )
        if response.status_code != 204:
            msg = f"An error occured during data ingest. Status {response.status_code}, message: {response.text}"
            self.log.warning(msg)
            raise ObeliskError(msg)
        return response

    async def fetch_single_chunk(self, params: QueryParams) -> QueryResult:
        response = await self.http_get(
            f"{self.kind.root_url}/{params.dataset}/data/query", params=params.to_dict()
        )

        if response.status_code != 200:
            self.log.warning(f"Unexpected status code: {response.status_code}")
            raise ObeliskError(response.status_code, response.reason_phrase)

        try:
            js = response.json()
            return QueryResult.model_validate(js)
        except json.JSONDecodeError as e:
            msg = f"Obelisk response is not a JSON object: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg) from e
        except ValidationError as e:
            msg = f"Response cannot be validated: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg) from e

    async def query(self, params: QueryParams) -> list[Datapoint]:
        params.cursor = None
        result_set: list[Datapoint] = []
        result_limit = params.limit

        # Obelisk CORE does not actually stop emitting a cursor when done, limit serves as page limit

        while True:
            params.limit = min(self.page_limit, result_limit - len(result_set))
            result: QueryResult = await self.fetch_single_chunk(params)
            result_set.extend(result.items)
            params.cursor = result.cursor

            if len(result_set) >= result_limit or result.cursor is None:
                break

        return result_set

    async def query_time_chunked(
        self, params: ChunkedParams
    ) -> AsyncIterator[list[Datapoint]]:
        for chunk in params.chunks():
            yield await self.query(chunk)



import pytest
from obelisk.asynchronous import Obelisk

# Intentionally public client, restricted to public datasets.
client_id = "682c6c46604b3b3be35429df"
client_secret = "7136832d-01be-456a-a1fe-25c7f9e130c5"

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
async def test_fetch_demo_igent():
    consumer = Obelisk(client=client_id, secret=client_secret)
    result = await consumer.fetch_single_chunk(
        datasets=["612f6c39cbceda0ea9753d95"],
        metrics=["org.dyamand.types.common.Temperature::number"],
        from_timestamp=1740924034000,
        to_timestamp=1741100614258,
        limit=2,
    )

    assert len(result.items) == 2



import json
from datetime import datetime, timedelta
from math import floor
from typing import Any, Literal
from collections.abc import AsyncGenerator

import httpx
from pydantic import ValidationError

from obelisk.exceptions import ObeliskError
from obelisk.types import Datapoint, IngestMode, QueryResult, TimestampPrecision

from obelisk.asynchronous.base import BaseClient


class Obelisk(BaseClient):
    """
    Component that contains all the logic to consume data from
    the Obelisk API (e.g. historical data, sse).

    For most usecases, `query` will be the method you need.
    Have a look at `query_time_chunked` too, because it might just be very useful.

    Obelisk API Documentation:
    https://obelisk.docs.apiary.io/
    """

    async def fetch_single_chunk(
        self,
        datasets: list[str],
        metrics: list[str] | None = None,
        fields: list[str] | None = None,
        from_timestamp: int | None = None,
        to_timestamp: int | None = None,
        order_by: dict[str, Any] | None = None,
        filter_: dict[str, Any] | None = None,
        limit: int | None = None,
        limit_by: dict[str, Any] | None = None,
        cursor: str | None = None,
    ) -> QueryResult:
        """
        Queries one chunk of events from Obelisk for given parameters,
        does not handle paging over Cursors.

        Parameters
        ----------

        - datasets:
            List of Dataset IDs.
        - metrics:
            List of Metric IDs or wildcards (e.g. `*::number`), defaults to all metrics.
        - fields:
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        - from_timestamp:
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        - to_timestamp:
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        - order_by:
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        - filter_:
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        - limit:
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        - limit_by:
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        - cursor:
            Specifies the next cursor,
            used when paging through large result sets.
        """

        # pylint: disable=too-many-arguments
        data_range = {"datasets": datasets}
        if metrics is not None:
            data_range["metrics"] = metrics

        payload = {
            "dataRange": data_range,
            "cursor": cursor,
            "fields": fields,
            "from": from_timestamp,
            "to": to_timestamp,
            "orderBy": order_by,
            "filter": filter_,
            "limit": limit,
            "limitBy": limit_by,
        }
        response = await self.http_post(
            self.kind.query_url,
            data={k: v for k, v in payload.items() if v is not None},
        )
        if response.status_code != 200:
            self.log.warning(f"Unexpected status code: {response.status_code}")
            raise ObeliskError(response.status_code, response.reason_phrase)

        try:
            js = response.json()
            return QueryResult.model_validate(js)
        except json.JSONDecodeError as e:
            msg = f"Obelisk response is not a JSON object: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg) from e
        except ValidationError as e:
            msg = f"Response cannot be validated: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg) from e

    async def query(
        self,
        datasets: list[str],
        metrics: list[str] | None = None,
        fields: list[str] | None = None,
        from_timestamp: int | None = None,
        to_timestamp: int | None = None,
        order_by: dict[str, Any] | None = None,
        filter_: dict[str, Any] | None = None,
        limit: int | None = None,
        limit_by: dict[str, Any] | None = None,
    ) -> list[Datapoint]:
        """
        Queries data from obelisk,
        automatically iterating when a cursor is returned.

        Parameters
        ----------

        - datasets:
            List of Dataset IDs.
        - metrics:
            List of Metric IDs or wildcards (e.g. `*::number`), defaults to all metrics.
        - fields:
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        - from_timestamp:
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        - to_timestamp:
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        - order_by:
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        - filter_:
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        - limit:
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        - limit_by:
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        """

        cursor: str | None | Literal[True] = True
        result_set: list[Datapoint] = []

        while cursor:
            actual_cursor = cursor if cursor is not True else None
            result: QueryResult = await self.fetch_single_chunk(
                datasets=datasets,
                metrics=metrics,
                fields=fields,
                from_timestamp=from_timestamp,
                to_timestamp=to_timestamp,
                order_by=order_by,
                filter_=filter_,
                limit=limit,
                limit_by=limit_by,
                cursor=actual_cursor,
            )
            result_set.extend(result.items)
            cursor = result.cursor

            if limit and len(result_set) >= limit:
                """On Obelisk HFS, limit is actually page size,
                so continuing to read the cursor will result in a larger than desired
                set of results.

                On the other hand, if the limit is very large,
                we may need to iterate before we reach the desired limit after all.
                """
                break

        return result_set

    async def query_time_chunked(
        self,
        datasets: list[str],
        metrics: list[str],
        from_time: datetime,
        to_time: datetime,
        jump: timedelta,
        filter_: dict[str, Any] | None = None,
        direction: Literal["asc", "desc"] = "asc",
    ) -> AsyncGenerator[list[Datapoint], None]:
        """
        Fetches all data matching the provided filters,
        yielding one chunk at a time.
        One "chunk" may require several Obelisk calls to resolve cursors.
        By necessity iterates over time, no other ordering is supported.

        Parameters
        ----------

        - datasets:
            Dataset IDs to query from
        - metrics:
            IDs of metrics to query
        - from_time:
            Start time to fetch from
        - to_time:
            End time to fetch until.
        - jump:
            Size of one yielded chunk
        - filter_:
            Obelisk filter, caller is responsible for correct format
        - direction:
            Yield older data or newer data first, defaults to older first.
        """

        current_start = from_time
        while current_start < to_time:
            yield await self.query(
                datasets=datasets,
                metrics=metrics,
                from_timestamp=floor(current_start.timestamp() * 1000),
                to_timestamp=floor((current_start + jump).timestamp() * 1000 - 1),
                order_by={"field": ["timestamp"], "ordering": direction},
                filter_=filter_,
            )
            current_start += jump

    async def send(
        self,
        dataset: str,
        data: list[dict[str, Any]],
        precision: TimestampPrecision = TimestampPrecision.MILLISECONDS,
        mode: IngestMode = IngestMode.DEFAULT,
    ) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        - dataset:
            ID for the dataset to publish to
        - data:
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.
        - precision:
            Precision used in the numeric timestamps contained in data.
            Ensure it matches to avoid weird errors.
        - mode:
            See docs for `obelisk.types.IngestMode`.

        Raises
        ------

        ObeliskError
            When the resulting status code is not 204, an empty `obelisk.exceptions.ObeliskError` is raised.
        """

        params = {
            "datasetId": dataset,
            "timestampPrecision": precision.value,
            "mode": mode.value,
        }

        response = await self.http_post(
            f"{self.kind.ingest_url}/{dataset}", data=data, params=params
        )
        if response.status_code != 204:
            msg = f"An error occured during data ingest. Status {response.status_code}, message: {response.text}"
            self.log.warning(msg)
            raise ObeliskError(msg)
        return response
